/* ------------------------------------
This is a stl-dependent version of a ROOT-dependent
program developed to post-process the output
generated by RasRD (stochdiff). Calculates the mean
concentration and the SD.
g++ -O2 -Wall basic00W.cpp -o basic00W
ex: ./basic00W HEK293model.out-cell2-conc.txt AMP
 ------------------------------------ */
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <sstream>
#include <cmath>
                    // provides sqrt() and pow()
#include <string>
#include <vector>
#include <iomanip>
using namespace std;

// ======================================
// Counts the number of words separated by a given token
// Args:
//     str        - the original string including tokens
//     delimiters - the delimiters that separate the words
// The template for this "tokenizer" has been taken from:
// http://www.oopweb.com/CPP/Documents/CPPHOWTO/Volume/C++Programming-HOWTO-7.html
// ======================================
int NTokens(const string& str, const string& delimiters = " ")
{
    int nmembers = 0;
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
        nmembers++;
    }
    return nmembers;
}
// ======================================
// Actually separate the words in a string
// Args:
//     str        - the original string including tokens
//     tokens     - vector to store the actual words
//     delimiters - the delimiters that separate the words
//  Used with the header
// The template for this "tokenizer" has been taken from:
// http://www.oopweb.com/CPP/Documents/CPPHOWTO/Volume/C++Programming-HOWTO-7.html
// ======================================
void Tokenizer(const string& str, vector<string>& tokens, const string& delimiters = " ")
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}
// ======================================
// Similar to Tokenizer but depends on the total
// number of members in the string (line).
// Args:
// M_member - a pointer to the 1d matrix "M_member"
// line     - the original string including tokens
// nmembers - the total number of elements in the string
// ======================================
int Token_matrix(vector<string>& M_member, const string& line,int nmembers)
                                                       // stores the strings in the 1d matrix  "M_member"
                                                       // receives: a pointer to the 1d matrix "M_member"
                                                       //           a Tstring line
                                                       //           a int nmembers

{
//        cout << "line: " << line << endl;
   string member;                                      // member stores each of ...
                                                       // ... the strings separated by " " in line
   M_member.clear();                                   // Clears the M_member (M_value vector)...
   stringstream strout;
   strout << line;
//        cout << "***** line: " << line << endl;
        for(int i = 0; i < nmembers; i++) {
           strout >> member;
//            cout << "member: "<< member << endl;
           M_member.push_back(member);                 // member is stored in M_member
//            cout << "M_member["<< i << "]: "<< M_member[i] << endl;
        }
   return 0;
}
// ======================================
//  Calculate the total number of compartments in
// the simulation based on information contained in the header
// ======================================
int Get_ncomps(vector<string>& M_member, const string& Region, const string& Specie, int nmembers)
{
   int ncomp               = 0;
   int SpecieLength        = Specie.length();
   string temp_member;
        for(int i = 0; i < nmembers; i++) {
        temp_member = M_member[i];
           if ( (temp_member.find("_"+Specie)!= string::npos) && (temp_member.find("_"+Region)!= string::npos) && (temp_member.substr(temp_member.length()-SpecieLength,temp_member.length())==Specie) ) {
                 ncomp++;
           }
        }
   return ncomp;
}
// ======================================
int Index_matrix(int *M_indexmember, vector<string>& M_member, int nmembers, const string& Region, const string& Specie)
{
   string member;
   int SpecieLength = Specie.length();
   int j            = 0;
   string temp_member;
        for(int i = 0; i < nmembers; i++) {
        temp_member = M_member[i];
          if ( (temp_member.find("_"+Specie)!= string::npos) && (temp_member.find("_"+Region)!= string::npos) &&
          (temp_member.substr(temp_member.length()-SpecieLength,temp_member.length())==Specie) ) {
           // cout << "M_indexmember["<<j<<"]: "
           // << i << endl;
           M_indexmember[j] = i;
           j++;
           }
        }
   return 0;
}
// ======================================
// --------------------------------------
int main (int argc, char* argv[])
{
// --------------------------------------                  // Get number of arguments passed to the program
    int argcROOT = argc;
// --------------------------------------
    cout << "argcROOT:    " << argcROOT << endl;           // shows it!

    vector<string> argvROOT;                               // declares a string pointer ...
//     argvROOT = new string [argcROOT];                   // ...to an array used to store the arguments
// --------------------------------------                  Uncomment to compile
   cout << argv[1] << endl;
   cout << argv[2] << endl;
   argvROOT.push_back(string(argv[1]));
   argvROOT.push_back(string(argv[2]));

// --------------------------------------
   cout << "argvROOT[0]: " << argvROOT[0] << endl;         // shows arg[0]
   cout << "argvROOT[1]: " << argvROOT[1] << endl;         // shows arg[1]

   char buffer[100];                                       //Buffer where to store the resulting formatted string.
   stringstream strout;
   ifstream     fin;
   ofstream     fout;
   string       Region    = "dendrite";
   // TString      SubRegion = "dendrite";                 // submembrane or cytosol
   string       Specie    = argvROOT[1];                   // stores the Specie name in "TString Specie"
   string       infname   = argvROOT[0];                   // stores the file name name in "TString infname"
   string       outfname  = argvROOT[0];                   // stores the file name name in "TString outfname"
   outfname.insert(outfname.find(".txt"),argvROOT[1]);     // insert Specie in outfname @ position before ".txt"
   cout << "infname:  " << infname  << endl;
   cout << "outfname: " << outfname << endl;
   int          nlines    = 0;                             // number of lines      in the file
   int          nmembers  = 0;                             // number of strings (members)
                                                           // spareted by " " in a given "line"
   int          ncomp     = 0;                             // number of ALL  subvolumes in the file
   int          ncompCYTO = 0;                             // number of CYTO subvolumes in the file
   int          ncompSUBM = 0;                             // number of SUBM subvolumes in the file
   int          k         = 0;                             // counter for CYTO subvolumes
   int          j         = 0;                             // counter for SUBM subvolumes
   int          *M_indexmember;                            // pointer to an array used to store indexes (ALL)
   int          *M_indexmemberCYTO;                        // pointer to an array used to store indexes (CYTO)
   int          *M_indexmemberSUBM;                        // pointer to an array used to store indexes (SUBM)
   float        sumconc;                                   // sum conc in all cytosolic subvolumes
   float        sumconcSD;                                 // sum conc in all SD subvolumes
   float        sumsquardconc;                             // squard sum conc in all cytosolic subvolumes
   float        sumsquardconcSD;                           // squard sum conc in all SD subvolumes
   float        meanconc;                                  // mean  conc in all cytosolic subvolumes
   float        meanconcSD;                                // mean  conc in all SD subvolumes
   float        varianceconc;                              // var   conc in all cytosolic subvolumes
   float        varianceconcSD;                            // var   conc in all SD subvolumes
   float        stdevconc;                                 // stdev conc in all cytosolic subvolumes
   float        stdevconc1;                                // stdev conc in all cytosolic subvolumes (alt)
   float        stdevconcSD;                               // stdev conc in all SD subvolumes
   float        stdevconcSD1;                              // stdev conc in all SD subvolumes (alt)


   string               line;                            // line stores a line @ a time
   string               member;
   vector<string>       M_value;                         // initialize the 1d string matrix (vector) of Values
   vector<string>       M_header;                        // initialize the 1d string matrix (vector) for Header


 // opens the file
   fin.open(infname.c_str());
   if (!fin.good()) {                                      // if it can't be opened...
     cout << "*****  Can't open: " << outfname << endl;    // ...issue a message
     return 1;
   }

//==============================================
   nmembers  = 0;
   getline(fin,line);                                         // reads the 1st line
   nmembers = NTokens(line);
                                                              // stores in "nmembers" ...
//                                                            // ...the number of strings (members)
//                                                            // spareted by " " in "line"
   cout << "nmembers: " << nmembers << endl;                  // shows this number (nmembers)


   Tokenizer(line, M_header);                                 // stores the header in the 1d matrix (vector) "M_header"

   ncomp     = Get_ncomps(M_header, Region, Specie,    \
   nmembers);                                                 // get "ncomps" (number of subvolumes) out of
   cout << "ncomps: "     << ncomp     << endl;               // M_header and shows this number (ncomps)
   ncompSUBM = Get_ncomps(M_header, "submembrane",     \
   Specie, nmembers);                                         // get "ncompsSUBM" (number of subvolumes) out of
   cout << "ncompsSUBM: " << ncompSUBM << endl;               // M_header and shows this number (ncomps)
   ncompCYTO = Get_ncomps(M_header, "cytosol",         \
   Specie, nmembers);                                         // get "ncompsCYTO" (number of subvolumes) out of
   cout << "ncompsCYTO: " << ncompCYTO << endl;               // M_header and shows this number (ncomps)



   M_indexmember      = new int [ncomp];                 // initialize the 1d string matrix of indexes (ALL)
   M_indexmemberSUBM  = new int [ncompSUBM];             // initialize the 1d string matrix of indexes (SUBM)
   M_indexmemberCYTO  = new int [ncompCYTO];             // initialize the 1d string matrix of indexes (CYTO)

   Index_matrix(M_indexmember, M_header, nmembers,       \
                Region, Specie);                         // fills matrix with indexes for subvolumes   (ALL)
   Index_matrix(M_indexmemberSUBM, M_header, nmembers,   \
                "submembrane", Specie);                  // fills matrix with indexes for subvolumes   (SUBM)
   Index_matrix(M_indexmemberCYTO, M_header, nmembers,   \
                "cytosol", Specie);                      // fills matrix with indexes for subvolumes   (CYTO)

   fout.open (outfname.c_str());
   fout << "#  " << "time"                    <<  " \t"  \
                 << "M_cytosol"+Specie        <<  " \t"  \
                 << "S_cytosol"+Specie        <<  " \t"  \
                 << "S_cytosol1"+Specie       <<  " \t"  \
                 << "M_submembrane"+Specie    <<  " \t"  \
                 << "S_submembrane"+Specie    <<  " \t"  \
                 << "S_submembrane1"+Specie   <<       endl;

// ---------------------------------------------
   fout.close();
//==============================================
   fout.open (outfname.c_str(), ios::app);
   fout.precision(3);
   fout.width(5);
      int intTokens = 0;
   while (1) {
      getline(fin,line);
      intTokens = NTokens(line);
      if (line.empty()) break;                                           // No new line

      Token_matrix(M_value, line, nmembers);

      if (nmembers!=intTokens)
        {
        cout << "*****  Incomplete line " << endl;
        break;                                           // Incomplete line
        }
      sumconc         = 0.0;
      sumconcSD       = 0.0;
      meanconc        = 0.0;
      meanconcSD      = 0.0;
      sumsquardconc   = 0.0;
      sumsquardconcSD = 0.0;
      stdevconc       = 0.0;
      stdevconc1      = 0.0;
      stdevconcSD     = 0.0;
      stdevconcSD1    = 0.0;
      varianceconc    = 0.0;
      varianceconcSD  = 0.0;
      j               =   0;
      k               =   0;

// ----------------------------------------- Calc Mean & Variance
      for (int i = M_indexmember[0]; i < (M_indexmember[0]+ncomp); i++){

         if (i==M_indexmemberSUBM[j]){
            sumconcSD = sumconcSD + atof(M_value[i].c_str()); //M_value[i].atof();
            sumsquardconcSD = sumsquardconcSD + pow(atof(M_value[i].c_str()),2); //pow(M_value[i].atof(),2);
            j++;
         }
         if (i==M_indexmemberCYTO[k]) {
            sumconc = sumconc + atof(M_value[i].c_str());
            sumsquardconc = sumsquardconc + pow(atof(M_value[i].c_str()),2);
            k++;
         }
      }
      meanconc     = sumconc/ncompCYTO;
      meanconcSD   = sumconcSD/ncompSUBM;
      j               =   0;
      k               =   0;
// ----------------------------------------- Calc variance (2nd loop)
      for (int i = M_indexmember[0]; i < (M_indexmember[0]+ncomp); i++){
         if (i==M_indexmemberSUBM[j]){
            varianceconcSD = varianceconcSD + pow(atof(M_value[i].c_str())-meanconcSD,2);
            j++;
         }
         if (i==M_indexmemberCYTO[k]) {
            varianceconc = varianceconc + pow(atof(M_value[i].c_str())-meanconc,2);
            k++;
         }
      }
// -----------------------------------------
      stdevconc    = sqrt(varianceconc/ncompCYTO);
      stdevconc1   = sqrt(sumsquardconc/ncompCYTO - pow(meanconc,2));
      stdevconcSD  = sqrt(varianceconcSD/ncompSUBM);
      stdevconcSD1 = sqrt(sumsquardconcSD/ncompSUBM - pow(meanconcSD,2));

      sprintf(buffer,"%07.03f",atof(M_value[0].c_str()));
      fout <<             buffer;
      sprintf(buffer,"%07.03f",meanconc);
      fout << " \t\t"  << buffer;
      sprintf(buffer,"%07.03f",stdevconc);
      fout << " \t\t"  << buffer;
      sprintf(buffer,"%07.03f",stdevconc1);
      fout << " \t\t"  << buffer;
      sprintf(buffer,"%07.03f",meanconcSD);
      fout << " \t\t"  << buffer;
      sprintf(buffer,"%07.03f",stdevconcSD);
      fout << " \t\t"  << buffer;
      sprintf(buffer,"%07.03f",stdevconcSD);
      fout << " \t\t"  << buffer  << endl;
    nlines++;
//     if (nlines==10) { break;}
   }
   fin.close();
   fout.close();
   return 0;
}
//==============================================
